# coding: utf-8
'''
ptrace constants, generated by test/gen.sh
'''

PTRACE_ATTACH = 16
PTRACE_CONT = 7
PTRACE_DETACH = 17
PTRACE_EVENT_CLONE = 3
PTRACE_EVENT_EXEC = 4
PTRACE_EVENT_EXIT = 6
PTRACE_EVENT_FORK = 1
PTRACE_EVENT_SECCOMP = 7
PTRACE_EVENT_VFORK = 2
PTRACE_EVENT_VFORK_DONE = 5
PTRACE_GETEVENTMSG = 16897
PTRACE_GETFPREGS = 14
PTRACE_GETREGS = 12
PTRACE_GETREGSET = 16900
PTRACE_GETSIGINFO = 16898
PTRACE_GETSIGMASK = 16906
PTRACE_INTERRUPT = 16903
PTRACE_KILL = 8
PTRACE_LISTEN = 16904
PTRACE_O_EXITKILL = 1048576
PTRACE_O_SUSPEND_SECCOMP = 2097152
PTRACE_O_TRACECLONE = 8
PTRACE_O_TRACEEXEC = 16
PTRACE_O_TRACEEXIT = 64
PTRACE_O_TRACEFORK = 2
PTRACE_O_TRACESECCOMP = 128
PTRACE_O_TRACESYSGOOD = 1
PTRACE_O_TRACEVFORK = 4
PTRACE_O_TRACEVFORKDONE = 32
PTRACE_PEEKDATA = 2
PTRACE_PEEKSIGINFO = 16905
PTRACE_PEEKTEXT = 1
PTRACE_PEEKUSER = 3
PTRACE_POKEDATA = 5
PTRACE_POKETEXT = 4
PTRACE_POKEUSER = 6
PTRACE_SEIZE = 16902
PTRACE_SETFPREGS = 15
PTRACE_SETOPTIONS = 16896
PTRACE_SETREGS = 13
PTRACE_SETREGSET = 16901
PTRACE_SETSIGINFO = 16899
PTRACE_SETSIGMASK = 16907
PTRACE_SINGLESTEP = 9
PTRACE_SYSCALL = 24
PTRACE_TRACEME = 0



# coding: utf-8

'''
Python wrapper for ptrace.
'''

import json
import ctypes
import platform
from collections import Sequence
import functools

from const import *
import _pyptrace

class X64UserRegs(ctypes.Structure):
    '''
    x64 UserRegs structure, see /usr/include/sys/user.h.
    '''
    _fields_ = [
        ('r15',     ctypes.c_uint64),
        ('r14',     ctypes.c_uint64),
        ('r13',     ctypes.c_uint64),
        ('r12',     ctypes.c_uint64),
        ('rbp',     ctypes.c_uint64),
        ('rbx',     ctypes.c_uint64),
        ('r11',     ctypes.c_uint64),
        ('r10',     ctypes.c_uint64),
        ('r9',      ctypes.c_uint64),
        ('r8',      ctypes.c_uint64),
        ('rax',     ctypes.c_uint64),
        ('rcx',     ctypes.c_uint64),
        ('rdx',     ctypes.c_uint64),
        ('rsi',     ctypes.c_uint64),
        ('rdi',     ctypes.c_uint64),
        ('orig_rax',ctypes.c_uint64),
        ('rip',     ctypes.c_uint64),
        ('cs',      ctypes.c_uint64),
        ('eflags',  ctypes.c_uint64),
        ('rsp',     ctypes.c_uint64),
        ('ss',      ctypes.c_uint64),
        ('fs_base', ctypes.c_uint64),
        ('gs_base', ctypes.c_uint64),
        ('ds',      ctypes.c_uint64),
        ('es',      ctypes.c_uint64),
        ('fs',      ctypes.c_uint64),
        ('gs',      ctypes.c_uint64),
    ]

class X32UserRegs(ctypes.Structure):
    '''
    x32 UserRegs structure, see /usr/include/sys/user.h.
    '''

    _fields_ = [
        ("ebx",			ctypes.c_uint32),
        ("ecx",			ctypes.c_uint32),
        ("edx",			ctypes.c_uint32),
        ("esi",			ctypes.c_uint32),
        ("edi",			ctypes.c_uint32),
        ("ebp",			ctypes.c_uint32),
        ("eax",			ctypes.c_uint32),
        ("xds",			ctypes.c_uint32),
        ("xes",			ctypes.c_uint32),
        ("xfs",			ctypes.c_uint32),
        ("xgs",			ctypes.c_uint32),
        ("orig_eax",	ctypes.c_uint32),
        ("eip",			ctypes.c_uint32),
        ("xcs",			ctypes.c_uint32),
        ("eflags",		ctypes.c_uint32),
        ("esp",			ctypes.c_uint32),
        ("xss",			ctypes.c_uint32),
    ]

class UnsupportArchException(Exception):
    pass

class PtraceException(Exception):
    pass

'''
UserRegs structure, PyPtrace will choose bwtween X32UserRegs/X64UserRegs
according to the machine arch that it is running on.
'''
UserRegs = None

arch = platform.machine()
if arch == 'x86_64':
    UserRegs = X64UserRegs
elif arch == 'i686':
    UserRegs = X32UserRegs
else:
    raise UnsupportArchException(arch)

class RegsWrapper(object):
    '''
    Wrapper for UserRegs, for pretty printing of UserRegs.
    '''

    def __init__(self, regs):
        self.regs = regs

    def __str__(self):
        if not self.regs:
            return None

        def reg_val(reg_name):
            reg = getattr(self.regs, reg_name)
            val = reg.real if hasattr(reg, 'real') else None
            if val: val = '0x{:016x}'.format(val)

            return val

        # register does not starts with '_'
        reg_names = [attr for attr in dir(self.regs) if not attr.startswith('_')]
        reg_dict = {reg_name: reg_val(reg_name) for reg_name in reg_names}
        return json.dumps(reg_dict, indent=4)

def check_ret(fn):
    '''
    Decorator for ptrace requests.
    This decorator will check the resutl for ptrace request and
    throw exception if throw_exception == True.
    '''

    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
        throw_exception = kwargs.get('throw_exception', True)
        if 'throw_exception' in kwargs:
            del kwargs['throw_exception']

        ret = fn(*args, **kwargs)
        errno = ret[0] if isinstance(ret, Sequence) else  ret 
        if errno != 0 and throw_exception is True:
            raise PtraceException('Failed executing %s' % fn.func_name)

        return ret 

    return wrapper

@check_ret
def attach(pid):
    '''
    Attach  to  the process specified in pid, making it a tracee of the
    calling process.
    '''

    return _pyptrace._pyptrace(PTRACE_ATTACH, pid, 0, 0)

@check_ret
def cont(pid, signo=0):
    '''
    Restart  the  stopped  tracee  process. If signo is nonzero, it is
    interpreted as the number of a signal to be delivered to the tracee;
    otherwise, no signal is delivered.
    '''

    return _pyptrace._pyptrace(PTRACE_CONT, pid, 0, signo)

@check_ret
def traceme():
    '''
    Indicate that this process is to be traced by its parent.
    '''

    return _pyptrace._pyptrace(PTRACE_TRACEME, 0, 0, 0)

@check_ret
def detach(pid, signo):
    '''
    Restart  the stopped tracee as for cont(), but first detach from it.
    '''

    return _pyptrace._pyptrace(PTRACE_DETACH, pid, 0, signo)

@check_ret
def peektext(pid, addr):
    '''
    Read  a  word  at  the address addr in the tracee's text memory, returning
    the word as the result of the peektext() call.
    '''

    return _pyptrace._pyptrace_peek(PTRACE_PEEKTEXT, pid, addr)

@check_ret
def peekdata(pid, addr):
    '''
    Read  a  word  at  the address addr in the tracee's data memory, returning
    the word as the result of the peektext() call.
    '''

    return _pyptrace._pyptrace_peek(PTRACE_PEEKDATA, pid, addr)

@check_ret
def peekuser(pid, addr):
    '''
    Read a word at offset addr in the tracee's USER area,  which  holds  the
    registers  and  other  information  about  the  process (see <sys/user.h>).
    '''

    return _pyptrace._pyptrace_peek(PTRACE_PEEKUSER, pid, addr)

@check_ret
def poketext(pid, addr, data):
    '''
    Copy the word data to the address addr in the tracee's text memory.
    '''

    return _pyptrace._pyptrace(PTRACE_POKETEXT, pid, addr, data)

@check_ret
def pokedata(pid, addr, data):
    '''
    Copy the word data to the address addr in the tracee's data memory.
    '''

    return _pyptrace._pyptrace(PTRACE_POKEDATA, pid, addr, data)

@check_ret
def pokeuser(pid, addr, data):
    '''
    Copy  the  word  data to offset addr in the tracee's USER area.
    '''

    return _pyptrace._pyptrace(PTRACE_POKEUSER, pid, addr, data)

@check_ret
def singlestep(pid, signo=0):
    '''
    Restart  the  stopped  tracee  as  for cont(), but arrange for the tracee
    to be stopped at the next entry to or exit after a single instruction.
    '''

    return _pyptrace._pyptrace(PTRACE_SINGLESTEP, pid, 0, signo)

@check_ret
def syscall(pid, signo=0):
    '''
    Restart  the  stopped  tracee  as  for cont(), but arrange for the tracee
    to be stopped at the next entry to or exit from a system call.
    '''

    return _pyptrace._pyptrace(PTRACE_SYSCALL, pid, 0, signo)

@check_ret
def setoptions(pid, options):
    '''
    Set  ptrace  options from options.
    '''

    return _pyptrace._pyptrace(PTRACE_SETOPTIONS, pid, 0, options)

_libc = ctypes.cdll.LoadLibrary('libc.so.6')

@check_ret
def getregs(pid):
    '''
    Return the tracee's general-purpose registers.
    '''

    regs = UserRegs()
    _libc_ptrace = _libc.ptrace
    _libc_ptrace.restype = ctypes.c_long
    _libc_ptrace.argtypes = (ctypes.c_long, ctypes.c_int,
                             ctypes.c_void_p, ctypes.POINTER(UserRegs))

    ret = _libc_ptrace(PTRACE_GETREGS, pid, None, ctypes.byref(regs))
    return ret, regs

@check_ret
def setregs(pid, regs):
    '''
    Modify the tracee's general-purpose registers,  respectively,  from  the
    paramater  regs.
    '''

    _libc_ptrace = _libc.ptrace
    _libc_ptrace.restype = ctypes.c_long
    _libc_ptrace.argtypes = (ctypes.c_long, ctypes.c_int,
                             ctypes.c_void_p, ctypes.POINTER(UserRegs))

    ret = _libc_ptrace(PTRACE_SETREGS, pid, None, ctypes.byref(regs))
    return ret



from setuptools import setup, find_packages, Extension

ptrace_module = Extension('_pyptrace',
    define_macros = [('MAJOR_VERSION', '1'),
        ('MINOR_VERSION', '7')],
    # include_dirs = ['/usr/local/include'],
    # libraries = ['pthread'],
    # library_dirs = ['/usr/local/lib'],
    sources=['pyptrace/_pyptrace.c'])

setup(
    name = 'pyptrace',
    description = 'Python wrapper for Linux ptrace system call.',
    author = 'wenlin.wu',
    author_email = 'wenlin.wu@outlook.com',
    url = 'https://github.com/kikimo/pyptrace',
    version = '1.7',
    packages = find_packages(),
    package_dir = {'':'.'},
    ext_modules = [ptrace_module],
    keywords = 'linux ptrace',
)
